// Oleg Sergeev, seoman@yandex.ru

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <ctime>

//Functions like $\int (ax+b)\cos (\omega x) dx$.
	double IntegrCos(double a, double b, double w, double x1, double x2)
	{
		double wsq = w*w;
		return ((a*cos(w*x2)/wsq + a*x2*sin(w*x2)/w + b*sin(w*x2)/w) - (a*cos(w*x1)/wsq + a*x1*sin(w*x1)/w + b*sin(w*x1)/w)); 
	}

	double IntegrSin(double a, double b, double w, double x1, double x2)
	{
		double wsq = w*w;
		return ((a*sin(w*x2)/wsq - a*x2*cos(w*x2)/w - b*cos(w*x2)/w) - (a*sin(w*x1)/wsq - a*x1*cos(w*x1)/w - b*cos(w*x1)/w)); 
	}

//The basis of a data structure is a two-way list of elements of "vectorlist" type, 
//to each of elements list of "vectors" is attached as well as intf and dynsf values array.

	struct vector {int x; int y; int z; int d2; double *ReStatFArray;
	double *ImStatFArray; vector *next;};
	struct vectorlist {int n; int d2; vector *vectors; double *ReIntFArray; double *ImIntFArray; double *ReDynFArray;
	double *ImDynFArray; vectorlist *prev; vectorlist *next;};

	vectorlist *InsertVector(int x, int y, int z, vectorlist *initlist)
	{
	int d2, trigger;
	vectorlist *curr, *innerinitlist, *prev, *next;
	vector *vect, *nextvect;
	d2 = x*x + y*y + z*z;
	trigger = 0;
	if (initlist == NULL)
	{	//Create new list and attach vector to it.
	innerinitlist = (vectorlist *)malloc(sizeof(vectorlist));
	curr = innerinitlist;
	(curr->prev) = NULL;
	(curr->next) = NULL;
	(curr->n) = 1;
	(curr->d2) = d2;
	vect = (vector *)malloc(sizeof(vector));
	(curr->vectors) = vect;
	(vect->x) = x;
	(vect->y) = y;
	(vect->z) = z;
	(vect->next) = NULL;
	}
	else	//if list exists
	{
	innerinitlist = initlist;
	curr = initlist;
	while (trigger == 0)	//will be in cycle until vector is inserted
	{
	if (curr == NULL)	//If we're at the end of the list
		{	//add new element
		curr = (vectorlist *)malloc(sizeof(vectorlist));
		(curr->prev) = prev;
		(prev->next) = curr;
		(curr->next) = NULL;
		(curr->n) = 1;
		(curr->d2) = d2;
		vect = (vector *)malloc(sizeof(vector));
		(curr->vectors) = vect;
		(vect->x) = x;
		(vect->y) = y;
		(vect->z) = z;
		(vect->next) = NULL;
		trigger = 1; //exit flag
		}
	else
	{
	if (d2 < (curr->d2))
		{	//insert new element before next, attach vector to new element
		next = curr;
		curr = (vectorlist *)malloc(sizeof(vectorlist));
		(curr->prev) = (next->prev);
		if ((curr->prev) != NULL)
		{(prev->next) = curr;}
		else
		{innerinitlist = curr;}
		(next->prev) = curr;
		(curr->next) = next;
		(curr->n) = 1;
		(curr->d2) = d2;
		vect = (vector *)malloc(sizeof(vector));
		(curr->vectors) = vect;
		(vect->x) = x;
		(vect->y) = y;
		(vect->z) = z;
		(vect->next) = NULL;
		trigger = 1; //exit flag
		}
	else
	if (d2 == (curr->d2))	//if the same length
		{	//just attach the vector
		(curr->n) = (curr->n)+1;
		nextvect = (curr->vectors);
		vect = (vector *)malloc(sizeof(vector));
		(curr->vectors) = vect;
		(vect->x) = x;
		(vect->y) = y;
		(vect->z) = z;
		(vect->next) = nextvect;
		trigger = 1; //exit flag
		}
	else
	if (d2 > (curr->d2))	//if didn't get to the place yet
		{	//keep in the loop
		prev = curr;
		curr = (curr->next);
		}
	}
	}
	}
	return innerinitlist; //return link to the list
	}

int main(int argc, char *argv[])
{

	FILE    *inpfile,
		*logfile,
		*dumpfile,
		*outfile1,
		*outfile2;

	int 	i, j, l, m; //counters

	vector *currvect, *prevvect, *nextvect;
	vectorlist *currlist, *prevlist, *nextlist, *initlist;
	
	time_t rawtime;
	
	int     x, y, z, d2,
		count,
		countvect,
		countlarge,
		countsmall,
		b,
		N,
		Tau,
		NumOfSteps,
		NonEquilSteps,
		NumOfPart,
		NumOfDumps,
		DumpStep,
		NumOfVect,
		MinInterval,
		MaxInterval,
		NumOfOmega;

	double	kx, ky, kz,
		Rx, Ry, Rz,
		temp,
		length,
		Omega,
		Time,
		BoxSize,
		VectStep,
		MinLthOfVect,
		MaxLthOfVect, //set
		MaxVect, //calculated
		TimeStep,
		InitOmega,
		OmegaStep,
		MaxOmega,
		ReIF,
		ImIF,
		ReDF,
		ImDF,
		ReCoupl,
		ImCoupl,
		ReSigma,
		ImSigma,
		ReTempExp,
		ImTempExp,
		*ReStatF,
		*ImStatF,
		*ReIntF,
		*ImIntF,
		*ReDisp,
		*ImDisp,
		*ReDynF,
		*ImDynF;

	char 	tmp[60], logstr[60], inpdata[60], fstr[60], outstr[60];   //temp strings


	if (argc != 2)
	{
		fprintf(stderr, "Incorrect usage of %s!\n", argv[0]);
		exit(1);
	}

//----------------------------------------------------------------------------

	if ((inpfile = fopen(argv[1], "r")) == NULL)
		{
			fprintf(stderr, "Can't open the file \"%s\"\n", argv[1]);
			exit(1);
		}

	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%d\n", &NumOfPart);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%le\n", &BoxSize);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%d\n", &NumOfSteps);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%d\n", &NonEquilSteps);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%le\n", &TimeStep);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%d\n", &DumpStep);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%le\n", &MinLthOfVect);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%le\n", &MaxLthOfVect);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%le\n", &OmegaStep);
	
//----------------------------------------------------------------------------

	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%s\n", logstr);

	if ((logfile = fopen(logstr, "w")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", logstr);
		exit(1);
	}
	time (&rawtime);
	fprintf(logfile, "Program started at %s\n", ctime(&rawtime));
	fclose(logfile);
	
//----------------------------------------------------------------------------

	initlist = NULL;
	length = 0;
	
//----------------------------------------------------------------------------

	if (MinLthOfVect > MaxLthOfVect)
	{
		fprintf(stderr, "Minimal length of wavevectors shouldn't be larger than the maximal one.\n");
		exit(1);
	}

	VectStep = 2*M_PI/BoxSize;
	MaxVect = 2*M_PI/(BoxSize/exp((log((double)NumOfPart))/3));
	if (MaxVect < MaxLthOfVect)
		MaxLthOfVect = MaxVect;

	NonEquilSteps = NonEquilSteps/DumpStep;
	NumOfDumps = NumOfSteps/DumpStep - NonEquilSteps + 1;
	
	MinInterval = 0;
	if (NumOfDumps <= 2000)
	MaxInterval = NumOfDumps/2;
	else
	MaxInterval = 1000;

	InitOmega = 2*M_PI/(MaxInterval*DumpStep*TimeStep);
	MaxOmega = 2*M_PI/(DumpStep*TimeStep);
	NumOfOmega = (int)((MaxOmega-InitOmega)/OmegaStep);

//----------------------------------------------------------------------------
//Form list of vectors, put it in file:

	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%s\n", tmp);
	
	if ((outfile1 = fopen(tmp, "w")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", tmp);
		exit(1);
	}

	m = 0;
	d2 = 0;
	x = 0; y = 0; z = 0;
	countvect = 0; countlarge = 1;
	while (countvect != countlarge)
	{
	countvect = 0; countsmall = 0; countlarge = 0;
	m++;
	for (x = 0; x <= m; x++)
	for (y = 0; y <= (m-x); y++)
		{
		z = m - (x+y);
		d2 = x*x + y*y + z*z;
		countvect++;
		if (d2*VectStep*VectStep > MaxLthOfVect*MaxLthOfVect)
		{countlarge++;}
		else
			{
			if (d2*VectStep*VectStep < MinLthOfVect*MinLthOfVect)
			{countsmall++;}
			else
				{
				initlist = InsertVector(x, y, z, initlist);
				}
			}
		}
	fprintf(outfile1, "m = %d, vect = %d, small = %d, large = %d\n", m, countvect, countsmall, countlarge);
	}

	if (initlist == NULL)
	{
		fprintf(stderr, "There are no suitable wavevectors. Please set other boundaries.\n");
		exit(1);
	}
	
	fprintf(outfile1, "\nList of vectors:\n(Unit of length equals %le)\n\n", VectStep);
	count = 0;

	currlist = initlist;
	while (currlist != NULL)
	{
	fprintf(outfile1, "d^2 = %d\n", (currlist->d2));
	currvect = (currlist->vectors);
		while (currvect != NULL)
		{
		fprintf(outfile1, "%d  %d  %d\n", (currvect->x),
		(currvect->y), (currvect->z));
		count++;
		currvect = (currvect->next);
		}
	fprintf(outfile1, "n = %d\n\n", (currlist->n));
	currlist = (currlist->next);
	}
	
	fprintf(outfile1, "Total number of vectors = %d\n", count);
	
	fclose(outfile1);
	
//Finished output.

	if ((logfile = fopen(logstr, "a")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", logstr);
		exit(1);
	}
	time (&rawtime);
	fprintf(logfile, "Vectors created, start of processing at %s\n", ctime(&rawtime));
	fprintf(stdout, "Vectors created, start of processing at %s\n", ctime(&rawtime));
	fclose(logfile);

//Calculate rho_k(t)

	currlist = initlist;
	while (currlist != NULL)
	{
	currvect = (currlist->vectors);
		while (currvect != NULL)
		{
//Create arrays for rho_k
		ReStatF = (double*)calloc(NumOfDumps, sizeof(double));
		for (i = 1; i <= NumOfDumps; i++)
			ReStatF[i-1] = 0;
	
		ImStatF = (double*)calloc(NumOfDumps, sizeof(double));
		for (i = 1; i <= NumOfDumps; i++)
			ImStatF[i-1] = 0;
			
		(currvect->ReStatFArray) = ReStatF;
		(currvect->ImStatFArray) = ImStatF;

		currvect = (currvect->next);
		}
	currlist = (currlist->next);
	}

	if ((logfile = fopen(logstr, "a")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", logstr);
		exit(1);
	}
	time (&rawtime);
	fprintf(logfile, "StatF arrays are allocated at %s\n", ctime(&rawtime));
	fclose(logfile);

	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%s\n", inpdata);
	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%s\n", fstr);

	if ((dumpfile = fopen(inpdata, "r")) == NULL)
		{
			fprintf(stderr, "Can't open the file \"%s\"\n", inpdata);
			exit(1);
		}

	for (l = 1; l <= NonEquilSteps; l++)
		{
			for (j = 1; j <= (9 + NumOfPart); j++) //9 strings in dump before coordinates
			{
			fgets(tmp, 60, dumpfile);
			}
		}
	for (l = 1; l <= NumOfDumps; l++)
	{
	for (j = 1; j <= 9; j++)
		{
		fgets(tmp, 60, dumpfile);
		}
	for (m = 1; m <= NumOfPart; m++)
		{
		fscanf(dumpfile, "%le %le %le\n", &Rx, &Ry, &Rz);
		currlist = initlist;
		while (currlist != NULL)
		{
		currvect = (currlist->vectors);
			while (currvect != NULL)
			{
				kx = (currvect->x)*VectStep;
				ky = (currvect->y)*VectStep;
				kz = (currvect->z)*VectStep;
				(currvect->ReStatFArray)[l-1] = (currvect->ReStatFArray)[l-1] + cos(kx*Rx + ky*Ry +kz*Rz);
				(currvect->ImStatFArray)[l-1] = (currvect->ImStatFArray)[l-1] + sin(kx*Rx + ky*Ry +kz*Rz);
				
			currvect = (currvect->next);
			}
		currlist = (currlist->next);
		}
		}
	}

	if ((logfile = fopen(logstr, "a")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", logstr);
		exit(1);
	}
	time (&rawtime);
	fprintf(logfile, "StatFs are calculated, dumpfile closed at %s\n", ctime(&rawtime));
	fclose(logfile);
	
	fclose(dumpfile);

	fgets(tmp, 60, inpfile);
	fscanf(inpfile, "%s\n", outstr);
	
	fclose(inpfile);


	if ((outfile2 = fopen(fstr, "w")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", fstr);
		exit(1);
	}
	
	fclose(outfile2);

	if ((outfile2 = fopen(outstr, "w")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", outstr);
		exit(1);
	}
	
	fprintf(outfile2, "k  Omega  Re(S)  Im(S)\n\n");
	fclose(outfile2);

	currlist = initlist;
	while (currlist != NULL)
	{
        ReIntF = (double*)calloc(MaxInterval+1, sizeof(double));
        for (i = 0; i <= MaxInterval; i++)
                ReIntF[i] = 0;

        ImIntF = (double*)calloc(MaxInterval+1, sizeof(double));
        for (i = 0; i <= MaxInterval; i++)
                ImIntF[i] = 0;

	(currlist->ReIntFArray) = ReIntF;
	(currlist->ImIntFArray) = ImIntF;

	ReDynF = (double*)calloc(NumOfOmega+1, sizeof(double));
	for (i = 1; i <= NumOfOmega; i++)
		ReDynF[i-1] = 0;		

	ImDynF = (double*)calloc(NumOfOmega+1, sizeof(double));
	for (i = 1; i <= NumOfOmega; i++)
		ImDynF[i-1] = 0;

	(currlist->ReDynFArray) = ReDynF;
	(currlist->ImDynFArray) = ImDynF;

	currvect = (currlist->vectors);
	while (currvect != NULL)
	{
		for (j = MinInterval; j <= MaxInterval; j++)
		{
			ReIF = 0;
			ImIF = 0;
			for(m = 1; m <= (NumOfDumps-j); m++)
			{
				ReCoupl = (currvect->ReStatFArray)[m-1]*(currvect->ReStatFArray)[m-1+j] + 
				(currvect->ImStatFArray)[m-1]*(currvect->ImStatFArray)[m-1+j];
				ImCoupl = (currvect->ReStatFArray)[m-1]*(currvect->ImStatFArray)[m-1+j] - 
				(currvect->ImStatFArray)[m-1]*(currvect->ReStatFArray)[m-1+j];
				ReCoupl = ReCoupl/NumOfPart;
				ImCoupl = ImCoupl/NumOfPart;
				ReIF = ReIF + ReCoupl;
				ImIF = ImIF + ImCoupl;
			}
			ReIF = ReIF/(NumOfDumps-j);
			ImIF = ImIF/(NumOfDumps-j);

                        (currlist->ReIntFArray)[j] = (currlist->ReIntFArray)[j] + ReIF;
                        (currlist->ImIntFArray)[j] = (currlist->ImIntFArray)[j] + ImIF;
		}
			
		currvect = (currvect->next);
	}
	
	for (j = MinInterval; j <= MaxInterval; j++)
	{
		(currlist->ReIntFArray)[j] = (currlist->ReIntFArray)[j]/(double)(currlist->n);
		(currlist->ImIntFArray)[j] = (currlist->ImIntFArray)[j]/(double)(currlist->n);
	}

	ReTempExp = 0;
	ImTempExp = 0;

	Omega = InitOmega;
	for (j = 1; j <= NumOfOmega; j++)
	{
			ReDF = 0;
			ImDF = 0;
			//Integrating from -tmax to tmax
			for (m=-MaxInterval; m < MaxInterval; m++)
			{
				//Values of F(k,t) in two succesive steps m and m+1 for numerical integration
				//including parity of F(k,t) into account.
				double ReF0,ImF0;
				if (m>=0) 
				{ 
					ReF0 = (currlist->ReIntFArray)[m]; 
					ImF0 = (currlist->ImIntFArray)[m];
				}
				else 
				{
					ReF0 = (currlist->ReIntFArray)[-m];
					ImF0 = (currlist->ReIntFArray)[-m];
				}
				
				double ReF1,ImF1;
				if (m+1>=0) 
				{ 
					ReF1 = (currlist->ReIntFArray)[m+1]; 
					ImF1 = (currlist->ImIntFArray)[m+1];
				}
				else 
				{
					ReF1 = (currlist->ReIntFArray)[-m-1];
					ImF1 = (currlist->ReIntFArray)[-m-1];
				}

				double ReFa = (ReF1 - ReF0)/(DumpStep*TimeStep);
				double ReFb = (ReF0*(double)(m+1) - ReF1*(double)m);
				double ImFa = (ImF1 - ImF0)/(DumpStep*TimeStep);
				double ImFb = (ImF0*(double)(m+1) - ImF1*(double)m);
				double time1 = (double)m*DumpStep*TimeStep;
				double time2 = (double)(m+1)*DumpStep*TimeStep;
				ReDF += IntegrCos(ReFa,ReFb,Omega,time1,time2) - IntegrSin(ImFa,ImFb,Omega,time1,time2);
				ImDF += IntegrSin(ReFa,ReFb,Omega,time1,time2) + IntegrCos(ImFa,ImFb,Omega,time1,time2);
			}	
			ReDF = ReDF * DumpStep * TimeStep/(2*M_PI);
			ImDF = ImDF * DumpStep * TimeStep/(2*M_PI);
			(currlist->ReDynFArray)[j-1] = ReDF;
			(currlist->ImDynFArray)[j-1] = ImDF;
			Omega = Omega + OmegaStep;
	}

	if ((outfile2 = fopen(fstr, "a")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", fstr);
		exit(1);
	}
	
	length = sqrt((double)(currlist->d2))*VectStep;
	for (j = MinInterval; j <= MaxInterval; j++)
	{
		fprintf(outfile2, "%le  %le  %le  %le\n", length, (double)j * DumpStep * TimeStep, (currlist->ReIntFArray)[j], (currlist->ImIntFArray)[j]);
	}
	fprintf(outfile2, "\n\n\n");
	fclose(outfile2);

	if ((outfile2 = fopen(outstr, "a")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", outstr);
		exit(1);
	}
	
	length = sqrt((double)(currlist->d2))*VectStep;
	Omega = InitOmega;
	for (j = 1; j <= NumOfOmega; j++)
	{
		fprintf(outfile2, "%le  %le  %le  %le\n", length,
		Omega, (currlist->ReDynFArray)[j-1], (currlist->ImDynFArray)[j-1]);
		Omega = Omega + OmegaStep;
	}
	fprintf(outfile2, "\n\n\n");
	fclose(outfile2);

	currlist = (currlist->next);
	}

	if ((logfile = fopen(logstr, "a")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", logstr);
		exit(1);
	}
	time (&rawtime);
	fprintf(logfile, "Output is finished at %s\n", ctime(&rawtime));
	fclose(logfile);

	//Memory cleaning
	currlist = initlist;
	while (currlist != NULL)
	{
		currvect = (currlist->vectors);
		while (currvect != NULL)
		{
		nextvect = (currvect->next);
		free(currvect->ReStatFArray);
		free(currvect->ImStatFArray);
		free(currvect);
		currvect = nextvect;
		}
	nextlist = (currlist->next);
	free(currlist->ReDynFArray);
	free(currlist->ImDynFArray);
	free(currlist);
	currlist = nextlist;
	}

	if ((logfile = fopen(logstr, "a")) == NULL)
	{
		fprintf(stderr, "Can't open the file \"%s\"\n", logstr);
		exit(1);
	}
	time (&rawtime);
	fprintf(logfile, "Program work is finished at %s", ctime(&rawtime));
	fclose(logfile);	

return 0;
}
